/*
  Play WORDLE with an ESP32...
*/
#define CHECK true // set false if you don't want to check your input is in the dict
#define NUMBER 5 // Number of letters (don't change unless you can change the dictionary)

#include "FS.h"
#include "SPIFFS.h"
#define FORMAT_SPIFFS_IF_FAILED true
#include <TFT_eSPI.h>
#include <SPI.h>
#include "Free_Fonts.h"
#define TFT_WIDTH 135
#define TFT_HEIGHT 240
#define BUTTON_1 35
#define BUTTON_2 0
TFT_eSPI tft = TFT_eSPI(TFT_WIDTH, TFT_HEIGHT);

//const char wordFile[] = "/Dict-12478.txt";
const char wordFile[] = "/Dict-5757.txt";

char Wordle[NUMBER + 1];
char Word[NUMBER + 1];
char Found[NUMBER + 1];
byte Alphabet[27] = {2};
byte check[NUMBER + 1] = {0}; // 2 well placed, 1 misplaced, 0 not in the word
char val[4] = "-=+";

void initTFT() {
  // Prepare the display
  tft.begin();
  tft.fillScreen (TFT_BLACK);
  tft.setRotation(3);  // 0 & 2 Portrait. 1 & 3 landscape
  tft.setTextDatum(TC_DATUM);
  tft.fillScreen (TFT_BLACK);
  tft.setFreeFont(FF40);
  tft.setTextColor(TFT_WHITE);
  tft.drawString("WORDLE", TFT_HEIGHT / 2, TFT_WIDTH / 3);
  tft.setTextDatum(BR_DATUM);
  tft.setFreeFont(FSS9);
  tft.setTextColor(TFT_BLUE);
  tft.drawString("[by Lesept]", TFT_HEIGHT - 10, TFT_WIDTH - 10);
  delay(3000);
  tft.setRotation(0);
  tft.fillScreen (TFT_BLACK);
  tft.setTextDatum(TC_DATUM);
}

void splashscreen() {
  Serial.println("\nWORDLE on ESP32 (by Lesept)");
  Serial.println("Guess the WORDLE in 6 tries. Each guess must be a valid 5 letter word.");
  Serial.println("My answer is:");
  Serial.printf("   %c : letter is not in the WORDLE\n", val[0]);
  Serial.printf("   %c : letter exists but not here\n", val[1]);
  Serial.printf("   %c : letter is well positionned\n", val[2]);
  Serial.println("Good luck!\n");
  Serial.printf("Please enter a %d letter word.\n", NUMBER);
  delay(200);
}

void sofar() {
  // Summarize what we know so far...
  Serial.println("Here is what you know so far...");
  Serial.printf("Well placed: %s\n", Found);
  Serial.print("Not in the Wordle: ");
  for (byte i = 0; i < 26; ++i)
    if (Alphabet[i] == 1) Serial.print(char('A' + i));
  Serial.println();
}

bool readWord() {
  while (Serial.available() == 0);
  byte n = 0;
  while (n < NUMBER) {
    if (Serial.available() > 0) {
      char letter = Serial.read();
      if (letter == '?') { // Recap what we know
        sofar();
        return false;
      }
      if (((letter >= 'a') && (letter <= 'z')) || ((letter >= 'A') && (letter <= 'Z'))) {
        if ((letter >= 'a') && (letter <= 'z')) letter = letter + 'A' - 'a';
        Word[n++] = letter;
      }
    }
  }
  Serial.println(Word);
  char t = Serial.read();
  while (t != '\n') t = Serial.read(); // Empty the buffer
  return true;
}

bool checkWord() {
  bool checkle[NUMBER + 1];
  byte nok = 0;

  // Check if letter is well placed
  for (byte i = 0; i < NUMBER; ++i) {
    checkle[i] = false;
    if (Wordle[i] == Word[i]) {
      check[i] = 2;
      Found[i] = Word[i];
      checkle[i] = true;
      ++ nok;
    } else check[i] = 0;
  }
  if (nok == NUMBER) return true; // All letters match!

  // Check if letter is in the word
  for (byte i = 0; i < NUMBER; ++i) {
    if (check[i] != 2) {
      for (byte j = 0; j < NUMBER; ++j) {
        if ((Word[i] == Wordle[j]) && !checkle[j] && check[j] != 2) {
          check[i] = 1;
          checkle[j] = true;
          break;
        }
      }
      bool isInWordle = false;
      for (byte j = 0; j < NUMBER; ++j)
        if (Word[i] == Wordle[j]) isInWordle = true;
      Alphabet[Word[i] - 'A'] = (isInWordle) ? 0 : 1;
    }
  }
  return false;
}

void pickRandom(File file) {
  // Pick a random word
  int pos = random(file.size() / (NUMBER + 1));
  file.seek(6 * pos, SeekSet);
  for (int i = 0; i < NUMBER; i++) {
    char letter = file.read();
    if ((letter >= 'a') && (letter <= 'z')) letter = letter + 'A' - 'a';
    Wordle[i] = letter;
  }
  file.close();
}


void setup ()
{
  pinMode(BUTTON_1, INPUT_PULLUP);
  pinMode(BUTTON_2, INPUT_PULLUP);
  Serial.begin (115200);
  initTFT();
  if (!SPIFFS.begin(FORMAT_SPIFFS_IF_FAILED)) {
    Serial.println("SPIFFS Mount Failed");
    return;
  }

  // Open the dictionary file
  File file = SPIFFS.open(wordFile);
  if (!file || file.isDirectory()) {
    Serial.printf("%s: Failed to open file for reading\n", wordFile);
    //  while (1);
  }
  splashscreen();
  for (byte i = 0; i < NUMBER; ++i) Found[i] = '-';
  pickRandom(file);
  //  Serial.println(Wordle); // The solution is here !

}

// Check if the proposed word is in the dictionary
bool exists() {
  char Test[NUMBER + 1];
  File file = SPIFFS.open(wordFile); // open dictionary
  int lines = file.size() / (NUMBER + 1);
  for (int i = 0; i < lines; ++i) {
    byte nok = 0;
    file.seek(6 * i, SeekSet);
    char letter = file.read(); // read first letter of each word
    if ((letter >= 'a') && (letter <= 'z')) letter = letter + 'A' - 'a';
    if (letter == Word[0]) { // if first letter matches
      nok = 1;;
      for (byte j = 1; j < NUMBER; ++j) { // read other letters
        letter = file.read();
        if ((letter >= 'a') && (letter <= 'z')) letter = letter + 'A' - 'a';
        if (letter == Word[j]) ++nok;
      }
      if (nok == NUMBER) {
        file.close();
        return true;
      }
    }
  }
  file.close();
  return false;
}

void dispWin(byte n) {
  tft.setRotation(3);
  tft.fillScreen (TFT_BLACK);
  tft.setTextDatum(TC_DATUM);
  tft.setFreeFont(FSS9);
  // tft.setCursor(20, 30);
  tft.setTextColor(TFT_RED);
  tft.drawString("CONGRATULATIONS !!!", TFT_HEIGHT / 2, 30);
  tft.println();
  tft.setFreeFont(FSS12);
  tft.setTextColor(TFT_GREEN);
  tft.drawString("WORDLE found", TFT_HEIGHT / 2, TFT_WIDTH / 2);
  char text[15];
  sprintf(text, "in %d trials", n + 1);
  tft.drawString(text, TFT_HEIGHT / 2, TFT_WIDTH / 2 + 30);
}

void dispFail() {
  tft.fillScreen (TFT_BLACK);
  tft.setTextDatum(TC_DATUM);
  tft.setFreeFont(FSS12);
  tft.setTextColor(TFT_BLUE);
  tft.drawString("The end !!!", TFT_WIDTH / 2, 27);
  tft.drawString("WORDLE", TFT_WIDTH / 2, 70);
  tft.drawString("was:", TFT_WIDTH / 2, 95);
  tft.setFreeFont(FSS18);
  tft.setTextColor(TFT_RED);
  tft.drawString(Wordle, TFT_WIDTH / 2, 150);
}

void dispNoDict() {
  tft.fillRect(0, TFT_HEIGHT - 63, TFT_WIDTH, 63, TFT_BLACK);
  tft.setTextDatum(TL_DATUM);
  tft.setFreeFont(FSS9);
  tft.setCursor(0, TFT_HEIGHT - 50);
  tft.setTextColor(TFT_RED);
  tft.print(Word);
  tft.setTextColor(TFT_BLUE);
  tft.println(" is not in");
  tft.println("my dictionary");
  tft.print("Please try again");
}

void dispCheck(int n)
{
  tft.setTextDatum(TC_DATUM);
  tft.setFreeFont(FSSB12);
  tft.fillRect(0, TFT_HEIGHT - 63, TFT_WIDTH, 63, TFT_BLACK);
  int h = 20;
  int d = 5;
  int y = n * (h + 3 * d);
  for (byte i = 0; i < NUMBER; ++i) {
    uint32_t color;
    if (check[i] == 2) color = TFT_GREEN;
    if (check[i] == 1) color = TFT_ORANGE;
    if (check[i] == 0) color = TFT_DARKGREY;
    int x = i * (h + d) + 3;
    tft.fillRect(x, y, h + d, h + d, color);
    tft.drawRect(x, y, h + d, h + d, TFT_BLUE);
    tft.setCursor(x + h / 4, y + h);
    tft.setTextColor(TFT_WHITE);
    tft.print(Word[i]);
  }
}

void loop() {
  static byte trials = 0; // only 6 trials
  if (readWord()) {// read the proposed word
    bool checkInput = true;
    if (CHECK) checkInput = exists(); // 450ms max
    if (!checkInput) { // test if this word is in the dictionary
      dispNoDict();
      Serial.println("This word is not in my dictionary, try again");
    } else {
      if (checkWord()) { // compute the score of the word
        dispWin(trials);
        Serial.printf("+++++\nCongratulations, you found it in %d trials!!", trials + 1);
        while (1);
      } else {
        dispCheck(trials);
        for (byte i = 0; i < NUMBER; ++i) Serial.print(val[check[i]]);
        Serial.println();
        if (++trials == NUMBER + 1) {
          dispFail();
          Serial.printf("THE END !!!\nThe solution was: %s\n", Wordle);
          while (1);
        }
      }
    }
  }
}
